from genologics.entities import Sample
from genologics.lims import Lims
from datetime import datetime as dt

from vogue.parse.build.sample import *


def build_sample(sample: Sample, lims: Lims, adapter) -> dict:
    """Build lims sample"""

    application_tag = sample.udf.get('Sequencing Analysis')
    category = adapter.get_category(application_tag)
    received_at = datetime2date(sample.udf.get('Received at'))
    delivered_at = datetime2date(sample.udf.get('Delivered at'))
    sequenced_at = datetime2date(sample.udf.get('Sequencing Finished'))
    prepared_at = datetime2date(sample.udf.get('Library Prep Finished'))

    mongo_sample = {'_id': sample.id}
    mongo_sample['sequenced_date'] = sequenced_at
    mongo_sample['received_date'] = received_at
    mongo_sample['prepared_date'] = prepared_at
    mongo_sample['delivery_date'] = delivered_at
    mongo_sample['sequenced_to_delivered'] = get_number_of_days(
        sequenced_at, delivered_at)
    mongo_sample['prepped_to_sequenced'] = get_number_of_days(
        prepared_at, sequenced_at)
    mongo_sample['received_to_prepped'] = get_number_of_days(
        received_at, prepared_at)
    mongo_sample['received_to_delivered'] = get_number_of_days(
        received_at, delivered_at)
    mongo_sample['family'] = sample.udf.get('Family')
    mongo_sample['strain'] = sample.udf.get('Strain')
    mongo_sample['source'] = sample.udf.get('Source')
    mongo_sample['customer'] = sample.udf.get('customer')
    mongo_sample['priority'] = sample.udf.get('priority')
    mongo_sample['initial_qc'] = sample.udf.get('Passed Initial QC')
    mongo_sample['library_qc'] = sample.udf.get('Passed Library QC')
    mongo_sample['prep_method'] = sample.udf.get('Prep Method')
    mongo_sample['sequencing_qc'] = sample.udf.get('Passed Sequencing QC')
    mongo_sample['application_tag'] = application_tag
    mongo_sample['category'] = category

    conc_and_amount = get_final_conc_and_amount_dna(application_tag, sample.id,
                                                    lims)
    mongo_sample['amount'] = conc_and_amount.get('amount')
    mongo_sample['amount-concentration'] = conc_and_amount.get('concentration')

    concentration_and_nr_defrosts = get_concentration_and_nr_defrosts(
        application_tag, sample.id, lims)
    mongo_sample['nr_defrosts'] = concentration_and_nr_defrosts.get(
        'nr_defrosts')
    mongo_sample[
        'nr_defrosts-concentration'] = concentration_and_nr_defrosts.get(
            'concentration')
    mongo_sample['lotnr'] = concentration_and_nr_defrosts.get('lotnr')

    mongo_sample[
        'microbial_library_concentration'] = get_microbial_library_concentration(
            application_tag, sample.id, lims)

    mongo_sample['library_size_pre_hyb'] = get_library_size(
        sample_id=sample.id, lims=lims, workflow='TWIST', size_steps=['KAPA Library Preparation TWIST v1'])
    mongo_sample['library_size_post_hyb'] = get_library_size(
        sample_id=sample.id, lims=lims, workflow='TWIST', size_steps=['Bead Purification TWIST v2', 'Bead Purification TWIST v1'])

    for key in list(mongo_sample.keys()):
        if mongo_sample[key] is None:
            mongo_sample.pop(key)

    return mongo_sample



#!/usr/bin/env python

import numpy as np


def find_concentration_amount(adapter, year: int = None) -> dict:
    """Prepares data for a scatter plot showning Concentration agains Amount."""

    amount = {"data": []}
    pipe = [
        {
            "$match": {
                "received_to_delivered": {"$exists": True},
                "amount": {"$exists": True},
                "amount-concentration": {"$exists": True},
            }
        },
        {"$project": {"year": {"$year": "$received_date"}, "amount": 1, "amount-concentration": 1}},
        {"$match": {"year": {"$eq": int(year)}}},
    ]

    aggregate_result = adapter.samples_aggregate(pipe)
    for sample in aggregate_result:
        if sample["amount"] > 200:
            sample["amount"] = 200
        amount["data"].append(
            {
                "x": sample["amount"],
                "y": round(sample["amount-concentration"], 2),
                "name": sample["_id"],
            }
        )
    return amount


def find_concentration_defrosts(adapter, year: int) -> dict:
    """Prepares data for a plot showning Number of defrosts against Concentration"""

    nr_defrosts = list(adapter.sample_collection.distinct("nr_defrosts"))
    defrosts = {
        "axis": {"x": "Number of Defrosts", "y": "Concentration (nM)"},
        "data": {},
        "title": "wgs illumina PCR-free",
        "labels": nr_defrosts.sort(),
    }

    pipe = [
        {
            "$match": {
                "received_to_delivered": {"$exists": True},
                "nr_defrosts-concentration": {"$exists": True},
                "nr_defrosts": {"$exists": True},
                "lotnr": {"$exists": True},
            }
        },
        {
            "$project": {
                "year": {"$year": "$received_date"},
                "nr_defrosts-concentration": 1,
                "nr_defrosts": 1,
                "lotnr": 1,
            }
        },
        {"$match": {"year": {"$eq": int(year)}}},
        {
            "$group": {
                "_id": {"nr_defrosts": "$nr_defrosts", "lotnr": "$lotnr"},
                "count": {"$sum": 1},
                "values": {"$push": "$nr_defrosts-concentration"},
            }
        },
        {"$sort": {"_id.nr_defrosts": 1}},
    ]
    aggregate_result = adapter.samples_aggregate(pipe)

    for result in aggregate_result:
        group = result["_id"]["lotnr"]
        if group not in defrosts["data"]:
            defrosts["data"][group] = {"median": [], "quartile": [], "nr_samples": []}
        nr = result["_id"]["nr_defrosts"]
        values = np.array(result["values"])
        defrosts["data"][group]["median"].append([nr, round(np.percentile(values, 50), 2)])
        defrosts["data"][group]["quartile"].append(
            [nr, round(np.percentile(values, 25), 2), round(np.percentile(values, 75), 2)]
        )
        defrosts["data"][group]["nr_samples"].append([nr, result["count"]])

    return defrosts



from vogue.exceptions import MissingApplicationTag


def build_application_tag(app_tag: dict) -> dict:
    """Builds the application tag collection documents.
    
    Args: 
        app_tag(dict): {'tag':'MELPCFR030', 'category':'wgs',...}
    Return:
        mongo_application_tag(dict): {'_id':'MELPCFR030', 'category':'wgs'} """

    tag = app_tag.get('tag')
    category = app_tag.get('prep_category')

    if not tag:
        raise MissingApplicationTag

    mongo_application_tag = {'_id': tag}
    if category:
        mongo_application_tag['category'] = category

    return mongo_application_tag



#!/usr/bin/env python

from vogue.constants.constants import MONTHS, YEARS, THIS_YEAR
from vogue.crud.find_plots.home import home_samples, home_customers
from vogue import __version__


from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter, templates

from fastapi import APIRouter, Depends, Request

router = APIRouter()


@router.get("/")
async def home(
    request: Request,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
    year: int = THIS_YEAR,
    month: int = 0,
):

    sample_series, cathegories = home_samples(adapter, int(year), month)
    customers = home_customers(adapter, int(year), month)
    return templates.TemplateResponse(
        "index.html",
        context={
            "request": request,
            "version": __version__,
            "sample_series": sample_series,
            "cathegories": cathegories,
            "customers": customers,
            "year_of_interest": year,
            "month_of_interest": month,
            "months": MONTHS,
            "month_name": MONTHS[month - 1][1] if month else "",
            "years": YEARS,
        },
    )



#!/usr/bin/env python


def home_samples(adapter, year, month):
    match = {"year": {"$eq": year}}
    if month:
        match["month"] = {"$eq": month}
    pipe = [
        {"$match": {"received_date": {"$exists": "True"}}},
        {
            "$project": {
                "month": {"$month": "$received_date"},
                "year": {"$year": "$received_date"},
                "category": 1,
                "priority": 1,
            }
        },
        {"$match": match},
        {
            "$group": {
                "_id": {"category": "$category", "priority": "$priority"},
                "count": {"$sum": 1},
            }
        },
    ]
    aggregate_result = adapter.samples_aggregate(pipe)
    samples = {}
    samples_output = {}
    cathegories = []
    for result in aggregate_result:
        cat = result["_id"].get("category", "missing")
        cathegories.append(cat)
        prio = result["_id"].get("priority", "missing")
        count = result.get("count", 0)
        if prio in samples:
            samples[prio][cat] = count
            continue
        samples[prio] = {cat: count}
        samples_output[prio] = []

    cathegories = list(set(cathegories))
    for prio, data in samples.items():
        for cat in cathegories:
            samples_output[prio].append(data.get(cat, 0))
    return samples_output, cathegories


def home_customers(adapter, year, month):
    match = {"year": {"$eq": year}}
    if month:
        match["month"] = {"$eq": month}
    pipe = [
        {"$match": {"received_date": {"$exists": "True"}}},
        {
            "$project": {
                "month": {"$month": "$received_date"},
                "year": {"$year": "$received_date"},
                "customer": 1,
            }
        },
        {"$match": match},
        {"$group": {"_id": {"customer": "$customer"}, "count": {"$sum": 1}}},
    ]
    aggregate_result = adapter.samples_aggregate(pipe)
    customers = {}
    for cust in aggregate_result:
        customer = cust["_id"].get("customer") if cust["_id"].get("customer") else "missing"
        customers[customer] = cust["count"]
    return customers



#!/usr/bin/env python
from typing import Literal

import numpy as np
from vogue.constants.constants import MONTHS


def microsalt_get_strain_st(adapter, year: int) -> dict:
    """Build aggregation pipeline to get information for microsalt sequence type vs strain data."""

    pipe = [
        {"$match": {"microsalt": {"$exists": "True"}}},
        {
            "$lookup": {
                "from": "sample",
                "localField": "_id",
                "foreignField": "_id",
                "as": "sample_info",
            }
        },
        {"$unwind": {"path": "$sample_info"}},
        {"$match": {"sample_info.received_date": {"$exists": "True"}}},
        {
            "$project": {
                "strain": "$sample_info.strain",
                "year": {"$year": "$sample_info.received_date"},
                "sequence_type": "$microsalt.blast_pubmlst.sequence_type",
            }
        },
        {"$match": {"year": {"$eq": int(year)}}},
        {
            "$group": {
                "_id": {"strain": "$strain", "sequence_type": "$sequence_type"},
                "number": {"$sum": 1},
            }
        },
        {
            "$match": {
                "number": {"$exists": "True"},
                "_id.sequence_type": {"$exists": "True", "$ne": ""},
            }
        },
        {
            "$group": {
                "_id": "$_id.strain",
                "number": {"$push": "$number"},
                "sequence_type": {"$push": "$_id.sequence_type"},
            }
        },
    ]

    aggregate_result = adapter.bioinfo_samples_aggregate(pipe)
    plot_data = {}
    for strain_results in aggregate_result:
        strain = strain_results["_id"]
        st = strain_results["sequence_type"]
        counts = strain_results["number"]
        data = [(st[i], c) for i, c in enumerate(counts)]
        plot_data[strain] = sorted(data)
    return plot_data


def microsalt_get_qc_time(
    adapter, year: int, metric_path: str, category: str, blast_pubmlst: Literal["Passed", "Failed"]
) -> dict:
    """Build aggregation pipeline to get information for microsalt qc data over time."""

    metric = metric_path.split(".")[1]
    pipe = [
        {"$match": {"microsalt.blast_pubmlst.thresholds": blast_pubmlst}},
        {
            "$lookup": {
                "from": "sample",
                "localField": "_id",
                "foreignField": "_id",
                "as": "sample_info",
            }
        },
        {"$unwind": {"path": "$sample_info"}},
        {
            "$match": {
                "sample_info.category": {"$eq": category},
                "sample_info.received_date": {"$exists": "True"},
            }
        },
        {
            "$project": {
                "month": {"$month": "$sample_info.received_date"},
                "year": {"$year": "$sample_info.received_date"},
                metric: "$microsalt." + metric_path,
            }
        },
        {"$match": {"year": {"$eq": int(year)}, metric: {"$not": {"$type": 2}}}},
        {
            "$group": {
                "_id": {"month": "$month"},
                metric: {"$push": f"${metric}"},
                "samples": {"$push": {metric: f"${metric}", "id": "$_id"}},
            }
        },
    ]
    aggregate_result = adapter.bioinfo_samples_aggregate(pipe)
    intermediate = {
        result["_id"]["month"]: {metric: result[metric], "samples": result["samples"]}
        for result in aggregate_result
    }
    box_plots = []
    labels = []
    means = []
    outliers = []
    for m in MONTHS:
        data = intermediate.get(m[0], [])
        if data:
            values = data[metric]
            samples = data["samples"]
            box_points = calculate_box_points(values)
            outliers_this_month = get_outliers(
                month=m[0] - 1,
                metric=metric,
                samples=samples,
                minimum=box_points[0],
                maximum=box_points[-1],
            )
            outliers.extend(outliers_this_month)
            means.append(np.mean(values))
            box_plots.append(box_points)
        else:
            box_plots.append([])
        labels.append(m[1])
    return {
        "outliers": outliers,
        "data": box_plots,
        "labels": labels,
        "mean": round(np.mean(means), 2),
    }


def get_outliers(month: int, metric: str, samples: list, minimum: float, maximum: float) -> list:
    """building scatter plot data for outlier samples a specific month. <metric> is what will be shown on the y axis"""

    outliers = []
    for sample in samples:
        if minimum > sample[metric] or sample[metric] > maximum:
            sample_data = {"x": month, "y": round(sample[metric], 2), "name": sample["id"]}
            outliers.append(sample_data)
    return outliers


def calculate_box_points(values: list) -> list:
    """calculate box points and outliers from a list of values"""

    first_quartile = round(np.percentile(values, 25), 2)
    second_quartile = round(np.percentile(values, 50), 2)
    third_quartile = round(np.percentile(values, 75), 2)
    inter_quartile_range = third_quartile - first_quartile
    maximum = third_quartile + inter_quartile_range * 1.5
    minimum = first_quartile - inter_quartile_range * 1.5
    return [
        round(minimum, 2),
        round(first_quartile, 2),
        round(second_quartile, 2),
        round(third_quartile, 2),
        round(maximum, 2),
    ]


def microsalt_get_untyped(adapter, year: int) -> dict:
    """Build aggregation pipeline to get information for microsalt qc data over time."""
    pipe = [
        {"$match": {"microsalt": {"$exists": "True"}}},
        {
            "$lookup": {
                "from": "sample",
                "localField": "_id",
                "foreignField": "_id",
                "as": "sample_info",
            }
        },
        {"$unwind": {"path": "$sample_info"}},
        {"$match": {"sample_info.received_date": {"$exists": "True"}}},
        {
            "$project": {
                "month": {"$month": "$sample_info.received_date"},
                "year": {"$year": "$sample_info.received_date"},
                "ST": "$microsalt.blast_pubmlst.thresholds",
            }
        },
        {"$match": {"year": {"$eq": int(year)}}},
        {"$group": {"_id": {"month": "$month"}, "ST": {"$push": "$ST"}}},
    ]

    intermediate_passed = {}
    intermediate_failed = {}
    aggregate_result = adapter.bioinfo_samples_aggregate(pipe)
    for result in aggregate_result:
        failed = 0
        passed = 0

        for sequence_type in result["ST"]:
            if sequence_type == "Failed":
                failed += 1
            elif sequence_type == "Passed":
                passed += 1
        intermediate_passed[result["_id"]["month"]] = passed
        intermediate_failed[result["_id"]["month"]] = failed

    data_passed = []
    data_failed = []
    labels = []
    for month in MONTHS:
        data_passed.append(intermediate_passed.get(month[0]))
        data_failed.append(intermediate_failed.get(month[0]))
        labels.append(month[1])

    return {"data_passed": data_passed, "data_failed": data_failed, "labels": labels}


def microsalt_get_st_time(adapter, year: int) -> dict:
    pipe = [
        {"$match": {"microsalt": {"$exists": "True"}}},
        {
            "$lookup": {
                "from": "sample",
                "localField": "_id",
                "foreignField": "_id",
                "as": "sample_info",
            }
        },
        {"$unwind": {"path": "$sample_info"}},
        {"$match": {"sample_info.received_date": {"$exists": "True"}}},
        {
            "$project": {
                "month": {"$month": "$sample_info.received_date"},
                "strain": "$sample_info.strain",
                "year": {"$year": "$sample_info.received_date"},
                "sequence_type": "$microsalt.blast_pubmlst.sequence_type",
            }
        },
        {
            "$match": {
                "year": {"$eq": int(year)},
                "sequence_type": {"$exists": "True"},
                "strain": {"$exists": "True"},
            }
        },
        {
            "$group": {
                "_id": {"month": "$month", "strain": "$strain", "sequence_type": "$sequence_type"},
                "number": {"$sum": 1},
            }
        },
    ]
    final_results = {}
    aggregate_result = list(adapter.bioinfo_samples_aggregate(pipe))
    for result in aggregate_result:
        st = result["_id"]["sequence_type"]
        strain = result["_id"]["strain"]
        if final_results.get(strain):
            final_results[strain][st] = [None] * 12
        else:
            final_results[strain] = {st: [None] * 12}
    for result in aggregate_result:
        count = result.get("number")
        if count:
            month = result["_id"]["month"]
            st = result["_id"]["sequence_type"]
            strain = result["_id"]["strain"]
            final_results[strain][st][month - 1] = count

    return {"data": final_results, "labels": [m[1] for m in MONTHS]}



import logging
from vogue.build.sample import build_sample

LOG = logging.getLogger(__name__)
from vogue.constants.constants import TEST_SAMPLES


def load_one(adapter, lims_sample=None, lims=None):
    """Function to load one lims sample into the database"""

    if lims_sample.id in TEST_SAMPLES:
        return

    mongo_sample = build_sample(lims_sample, lims, adapter)
    adapter.add_or_update_document(mongo_sample, adapter.sample_collection)


def load_all(adapter, lims, start_sample=None):
    """Function to load all lims samples into the database"""

    all_samples = lims.get_samples()
    total_nr_samples = len(all_samples)
    for nr, sample in enumerate(all_samples):
        if not start_sample:
            LOG.info("%s/%s %s" % (nr, total_nr_samples, sample.id))
            load_one(adapter, lims_sample=sample, lims=lims)
        elif start_sample and start_sample == sample.id:
            start_sample = None


def load_recent(adapter, lims, the_date):
    """Function to load all lims samples into the database"""

    latest_processes = lims.get_processes(last_modified=the_date)
    samples = []
    LOG.info("Found %s processes modified since %s.", len(latest_processes), the_date)
    LOG.info("Fetching recently updated samples...")
    for process in latest_processes:
        for analyte in process.all_inputs():
            samples += analyte.samples
    unique_samples = set(samples)
    nr_unique_samples = len(unique_samples)
    LOG.info("%s samples will be added or updated.", nr_unique_samples)
    for nr, sample in enumerate(unique_samples):
        LOG.info("%s/%s %s" % (nr, nr_unique_samples, sample.id))
        load_one(adapter, lims_sample=sample, lims=lims)


def load_one_dry(lims_sample, lims, adapter):
    existing_sample = adapter.sample(lims_sample.id)
    if existing_sample:
        LOG.info("The sample exists in the database.")
    else:
        LOG.info("The sample does not exists in the database.")
    mongo_sample = build_sample(lims_sample, lims, adapter)
    LOG.info("Sample informamtion from lims to add/update: \n %s", mongo_sample)
    return


def load_all_dry():
    LOG.info("Will load all lims samples.")
    return



from fastapi import FastAPI

from vogue.api.api_v1.endpoints import (
    insert_documents,
    home,
    common_trends,
    sequencing,
    genootype,
    reagent_labels,
    prepps,
    bioinfo_covid,
    bioinfo_micro,
    bioinfo_mip,
    update,
)
from vogue.settings import static_files


app = FastAPI()
app.mount(
    "/static",
    static_files,
    name="static",
)

app.include_router(home.router, tags=["home"])
app.include_router(common_trends.router, tags=["common_trends"])
app.include_router(sequencing.router, tags=["sequencing"])
app.include_router(genootype.router, tags=["genotype"])
app.include_router(reagent_labels.router, tags=["index"])
app.include_router(prepps.router, tags=["preps"])
app.include_router(bioinfo_micro.router, tags=["bioinfo_micro"])
app.include_router(bioinfo_covid.router, tags=["bioinfo_covid"])
app.include_router(bioinfo_mip.router, tags=["bioinfo_mip"])
app.include_router(update.router, tags=["update"])
app.include_router(insert_documents.router, tags=["sample"])



import numpy as np
from vogue.constants.constants import TEST_SAMPLES


def pipe_value_per_month(year: int, y_val: str, group_key: str = None) -> list:
    """Build aggregation pipeline to get information for a plot from the sample colection.

    A plot is going to show some average value (determined by y_val) per month,
    grouped by some group (determined by group_key). And it will only show
    data from a specific year (determined by year). If y_val='count' the value on the y-axis
    will instead be nr per samples.

    Arguments:
        year(int):      Any year.
        y_val(str):     A key (of a sample document) that we want to plot. Or 'count'
        group_key(str): A key (of the sample document) on wich we want to group. Can be None!

    eg 1:
        y_val = 'library_size_post_hyb'
        year = 2018
        group_key = 'source'

        Plot content:
            Average library_size_post_hyb/month for all samples during 2018 grouped by source

    eg 2:
        y_val = 'count'
        year = 2017
        group_key = 'priority'

        Plot content:
            Number of revieved samples/month during 2017 grouped by priority.
    """

    match = {"$match": {"received_date": {"$exists": True}, "_id": {"$nin": TEST_SAMPLES}}}
    project = {
        "$project": {"month": {"$month": "$received_date"}, "year": {"$year": "$received_date"}}
    }
    match_year = {"$match": {"year": {"$eq": year}}}
    group = {"$group": {"_id": {"month": "$month"}}}
    sort = {"$sort": {"_id.month": 1}}

    if group_key:  # grouping by group_key
        match["$match"][group_key] = {"$exists": True}
        project["$project"][group_key] = 1
        group["$group"]["_id"][group_key] = "$" + group_key
        sort["$sort"]["_id." + group_key] = 1

    if y_val == "count":
        # count nr samples per month:
        group["$group"][y_val] = {"$sum": 1}
    else:
        # get average of y_val:
        match["$match"][y_val] = {"$exists": True}
        project["$project"][y_val] = 1
        group["$group"][y_val] = {"$push": "$" + y_val}

    return [match, project, match_year, group, sort]


def value_per_month(adapter, year: int, y_val: str, group_key: str = None):
    """Wraper function that will build a pipe, aggregate results from database and
    prepare the data for the plot."""

    pipe = pipe_value_per_month(year, y_val, group_key)
    aggregate_result = adapter.samples_aggregate(pipe)
    return reformat_aggregate_results(list(aggregate_result), y_val, group_key)


def reformat_aggregate_results(aggregate_result, y_val, group_key=None):
    """Reformats raw output from the aggregation query to the format required by the plots.

    Arguments:
        aggregate_result (list): output from aggregation query.
        y_val(str):   A key (of a sample document) that we want to plot. Or 'count'
        group_key(str): A key (of the sample document) on wich we want to group. Can be None!
    Returns:
        plot_data(dict): see example

    Example:
        aggregate_result:
        [{'_id': {'strain': 'A. baumannii', 'month': 1}, 'microbial_library_concentration': 21.96},
         {'_id': {'strain': 'A. baumannii', 'month': 3}, 'microbial_library_concentration': 43.25},
         ...,
         {'_id': {'strain': 'E. coli', 'month': 2}, 'microbial_library_concentration': 7.68},
         ...]

        plot_data: {'A. baumannii': {'data': [21.96, None, 43.25,...]},
                    'E. coli': {'data': [None, 7.68, ...]},
                                            ...}
    """

    plot_data = {}
    for group in aggregate_result:
        if group_key:
            group_name = group["_id"][group_key]
        else:
            group_name = "all_samples"
        month = group["_id"]["month"]
        if y_val == "count":
            value = group[y_val]
        else:
            value = np.mean([float(val) for val in group[y_val]])

        if group_name not in plot_data:
            plot_data[group_name] = {"data": [None] * 12}
        plot_data[group_name]["data"][month - 1] = value
    return plot_data



def build_reagent_label_category(lims_reagent_label) -> dict:
    """Build reagent label ctegory document from lims data."""

    mongo_reagent_label = {
        '_id': lims_reagent_label.name,
        'sequence': lims_reagent_label.sequence,
        'category': lims_reagent_label.category,
        'name': lims_reagent_label.name
    }

    return mongo_reagent_label



#!/usr/bin/env python
from typing import List


from vogue.settings import templates

from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter
from fastapi import APIRouter, Depends, Request

router = APIRouter()
from vogue.constants.constants import (
    YEARS,
    BIOINFO_HELP_URLS,
    DNA_PICARD_SUB_SET,
    MONTHS,
    THIS_YEAR,
)
from vogue.crud.find_plots.bioinfo.qc import qc_dna_picard_plot, qc_dna_picard_time_plot
from vogue import __version__


@router.get("/Bioinfo/picard_time")
async def qc_dna_picard_time(
    request: Request,
    picard_metric: str = "PICARD_INSERT_SIZE MEAN_INSERT_SIZE",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    qc_dna_results: List[dict] = qc_dna_picard_time_plot(adapter, year)
    selected_group, selcted_metric = picard_metric.split()
    return templates.TemplateResponse(
        "qc_over_time.html",
        context=dict(
            request=request,
            selected_group=selected_group,
            selcted_metric=selcted_metric,
            qc_dna_results=qc_dna_results,
            DNA_PICARD=DNA_PICARD_SUB_SET,
            help_urls=BIOINFO_HELP_URLS,
            months=[m[1] for m in MONTHS],
            header="QC over time",
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )


@router.get(
    "/Bioinfo/picard",
)
async def qc_dna_picard(
    request: Request,
    Y_axis: str = "PICARD_INSERT_SIZE MEAN_INSERT_SIZE",
    X_axis: str = "PICARD_INSERT_SIZE MEAN_INSERT_SIZE",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    qc_dna_results: List[dict] = qc_dna_picard_plot(adapter, year)
    Y_group, Y_axis = Y_axis.split()
    X_group, X_axis = X_axis.split()
    return templates.TemplateResponse(
        "qc_dna_picard.html",
        context=dict(
            request=request,
            Y_axis=Y_axis,
            X_axis=X_axis,
            groups=list(set([Y_group, X_group])),
            qc_dna_results=qc_dna_results,
            DNA_PICARD=DNA_PICARD_SUB_SET,
            help_urls=BIOINFO_HELP_URLS,
            header="QC plots",
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )



"""
    cli for handling bioinfo collections. Addition and update!
"""
import logging
import click

# commands
from vogue.commands.load.bioinfo.bioinfo_raw import bioinfo_raw as bioinfo_raw_command
from vogue.commands.load.bioinfo.bioinfo_process import bioinfo_process as bioinfo_process_command
from vogue.commands.load.bioinfo.bioinfo_sample import bioinfo_sample as bioinfo_sample_command

# Get version and doc decorator
from vogue import __version__
from vogue.tools.cli_utils import add_doc as doc

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@doc("Vogue {version}: A trending package".format(version=__version__))
@click.pass_context
def bioinfo():
    """Main entry point of load commands"""

    pass


bioinfo.add_command(bioinfo_raw_command)
bioinfo.add_command(bioinfo_process_command)
bioinfo.add_command(bioinfo_sample_command)



import logging
import click
from vogue.load.reagent_label import load_one, load_all, load_recent
from datetime import date, timedelta


LOG = logging.getLogger(__name__)


@click.command("reagent_labels", short_help="Load reagent_labels into db.")
@click.option(
    "-a",
    "--all-reagent_labels",
    is_flag=True,
    help="Loads reagent_labels from all flowcells found in LIMS.",
)
@click.option("--dry", is_flag=True, help="Load reagent_labels from flowcell or not. (dry-run)")
@click.option(
    "-d",
    "--days",
    type=int,
    help="Update only reagent_labels from runs updated in the latest number of days",
)
@click.pass_context
def reagent_labels(ctx: click.Context, all_reagent_labels, dry, days):
    """Read and load lims data for a one or all many reagent_labels"""
    adapter = ctx.obj["adapter"]
    lims = ctx.obj["lims"]
    if not lims:
        LOG.warning("Lims connection failed.")
        raise click.Abort()

    if days:
        try:
            some_days_ago = date.today() - timedelta(days=days)
            the_date = some_days_ago.strftime("%Y-%m-%dT00:00:00Z")
        except Exception as err:
            LOG.error(err)
            raise click.Abort()
        load_recent(adapter, lims, the_date)
    elif all_reagent_labels:
        load_all(adapter, lims=lims)



#!/usr/bin/env python
import logging

import click
from genologics.lims import Lims

from vogue.settings import get_vogue_adapter, get_lims

LOG = logging.getLogger(__name__)

# commands
from vogue.commands.load.bioinfo import bioinfo as bioinfo_command
from vogue.commands.load.application_tag import application_tags as status_db_command
from vogue.commands.load.flowcell import flowcell as flowcell_command
from vogue.commands.load.sample import sample as sample_command
from vogue.commands.load.genotype import genotype as genotype_command
from vogue.commands.load.reagent_label import reagent_labels as reagent_label_command
from vogue.commands.load.reagent_label_category import (
    reagent_label_categories as reagent_label_category_command,
)

# Get version and doc decorator
from vogue import __version__
from vogue.tools.cli_utils import add_doc as doc


@click.group()
@click.version_option(version=__version__)
@click.pass_context
@doc("Vogue {version}: A trending package".format(version=__version__))
def load(ctx):
    """Main entry point of load commands"""

    ctx.ensure_object(dict)
    ctx.obj["lims"] = get_lims()
    ctx.obj["adapter"] = get_vogue_adapter()


load.add_command(bioinfo_command)
load.add_command(status_db_command)
load.add_command(sample_command)
load.add_command(flowcell_command)
load.add_command(genotype_command)
load.add_command(reagent_label_command)
load.add_command(reagent_label_category_command)



from typing import List

from genologics.lims import Lims
from genologics.entities import Sample, Artifact
from vogue.constants.lims_constants import MASTER_STEPS_UDFS
from datetime import time, datetime as dt
import operator
import logging

LOG = logging.getLogger(__name__)


def str_to_datetime(date: str) -> dt:
    """Convert str to datetime"""
    if date is None:
        return None
    return dt.strptime(date, '%Y-%m-%d')


def datetime2date(date: dt) -> dt.date:
    """Convert datetime.datetime to datetime.date"""

    if date is None:
        return None
    return dt(date.year, date.month, date.day)


def get_number_of_days(first_date: dt, second_date: dt) -> int:
    """Get number of days between different time stamps."""

    days = None
    if first_date and second_date:
        time_span = second_date - first_date
        days = time_span.days

    return days


def get_output_artifact(process_types: list,
                        lims_id: str,
                        lims: Lims,
                        last: bool = True) -> Artifact:
    """Returns the output artifact related to lims_id and the step that was first/latest run.
    
    If last = False return the first artifact
    """
    artifacts = lims.get_artifacts(samplelimsid=lims_id,
                                   process_type=process_types)
    artifact = None
    date = None
    for art in artifacts:
        # Get the date of the artifact
        new_date = str_to_datetime(art.parent_process.date_run)
        if not new_date:
            continue
        # If this is the first artifact we initialise the variables
        if not date:
            date = new_date
        if not artifact:
            artifact = art
            continue
        # If we want the latest artifact check if new date is newer than existing
        if last:
            if new_date > date:
                artifact = art
                date = new_date
        # If we want the first artifact check if new date is older than existing
        else:
            if new_date < date:
                artifact = art
                date = new_date

    return artifact


def get_latest_input_artifact(process_type: str, lims_id: str,
                              lims: Lims) -> Artifact:
    """Returns the input artifact related to lims_id and the step that was latest run."""
    latest_input_artifact = None
    artifacts = lims.get_artifacts(samplelimsid=lims_id,
                                   process_type=process_type)
    # Make a list of tuples (<date the artifact was generated>, <artifact>):
    date_art_list = list(
        set([(a.parent_process.date_run, a) for a in artifacts]))
    if date_art_list:
        # Sort on date:
        date_art_list.sort(key=operator.itemgetter(0))
        # Get latest:
        dummy, latest_outart = date_art_list[-1]  # get latest
        # Get the input artifact related to our sample
        for inart in latest_outart.input_artifact_list():
            if lims_id in [sample.id for sample in inart.samples]:
                latest_input_artifact = inart
                break

    return latest_input_artifact


def get_concentration_and_nr_defrosts(application_tag: str, lims_id: str,
                                      lims: Lims) -> dict:
    """Get concentration and nr of defrosts for wgs illumina PCR-free samples.

    Find the latest artifact that passed through a concentration_step and get its 
    concentration_udf. --> concentration
    Go back in history to the latest lot_nr_step and get the lot_nr_udf from that step. --> lotnr
    Find all steps where the lot_nr was used. --> all_defrosts
    Pick out those steps that were performed before our lot_nr_step --> defrosts_before_this_process
    Count defrosts_before_this_process. --> nr_defrosts"""

    if not application_tag:
        return {}

    if not application_tag[0:6] in MASTER_STEPS_UDFS[
        'concentration_and_nr_defrosts']['apptags']:
        return {}

    lot_nr_steps = MASTER_STEPS_UDFS['concentration_and_nr_defrosts'][
        'lot_nr_step']
    concentration_step = MASTER_STEPS_UDFS['concentration_and_nr_defrosts'][
        'concentration_step']
    lot_nr_udf = MASTER_STEPS_UDFS['concentration_and_nr_defrosts'][
        'lot_nr_udf']
    concentration_udf = MASTER_STEPS_UDFS['concentration_and_nr_defrosts'][
        'concentration_udf']

    return_dict = {}
    concentration_art = get_latest_input_artifact(concentration_step, lims_id,
                                                  lims)
    if concentration_art:
        concentration = concentration_art.udf.get(concentration_udf)
        lotnr = concentration_art.parent_process.udf.get(lot_nr_udf)
        this_date = str_to_datetime(concentration_art.parent_process.date_run)

        # Ignore if multiple lot numbers:
        if lotnr and len(lotnr.split(',')) == 1 and len(lotnr.split(' ')) == 1:
            all_defrosts = []
            for step in lot_nr_steps:
                all_defrosts += lims.get_processes(type=step,
                                                   udf={lot_nr_udf: lotnr})
            defrosts_before_this_process = []

            # Find the dates for all processes where the lotnr was used (all_defrosts),
            # and pick the once before or equal to this_date
            for defrost in all_defrosts:
                if defrost.date_run and str_to_datetime(
                        defrost.date_run) <= this_date:
                    defrosts_before_this_process.append(defrost)

            nr_defrosts = len(defrosts_before_this_process)

            return_dict = {
                'nr_defrosts': nr_defrosts,
                'concentration': concentration,
                'lotnr': lotnr,
                'concentration_date': this_date
            }

    return return_dict


def get_final_conc_and_amount_dna(application_tag: str, lims_id: str,
                                  lims: Lims) -> dict:
    """Find the latest artifact that passed through a concentration_step and get its 
    concentration. Then go back in history to the latest amount_step and get the amount."""

    if not application_tag:
        return {}

    if not application_tag[0:6] in MASTER_STEPS_UDFS[
        'final_conc_and_amount_dna']['apptags']:
        return {}

    return_dict = {}
    amount_udf = MASTER_STEPS_UDFS['final_conc_and_amount_dna']['amount_udf']
    concentration_udf = MASTER_STEPS_UDFS['final_conc_and_amount_dna'][
        'concentration_udf']
    concentration_step = MASTER_STEPS_UDFS['final_conc_and_amount_dna'][
        'concentration_step']
    amount_step = MASTER_STEPS_UDFS['final_conc_and_amount_dna']['amount_step']

    concentration_art = get_latest_input_artifact(concentration_step, lims_id,
                                                  lims)

    if concentration_art:
        amount_art = None
        step = concentration_art.parent_process
        # Go back in history untill we get to an output artifact from the amount_step
        while step and not amount_art:
            art = get_latest_input_artifact(step.type.name, lims_id, lims)
            processes = [
                p.type.name
                for p in lims.get_processes(inputartifactlimsid=art.id)
            ]
            for step in amount_step:
                if step in processes:
                    amount_art = art
                    break
            step = art.parent_process

        amount = amount_art.udf.get(amount_udf) if amount_art else None
        concentration = concentration_art.udf.get(concentration_udf)
        return_dict = {'amount': amount, 'concentration': concentration}

    return return_dict


def get_microbial_library_concentration(application_tag: str, lims_id: str,
                                        lims: Lims) -> float:
    """Check only samples with mictobial application tag.
    Get concentration_udf from concentration_step."""

    if not application_tag:
        return {}

    if not application_tag[3:5] == MASTER_STEPS_UDFS[
        'microbial_library_concentration']['apptags']:
        return None

    concentration_step = MASTER_STEPS_UDFS['microbial_library_concentration'][
        'concentration_step']
    concentration_udf = MASTER_STEPS_UDFS['microbial_library_concentration'][
        'concentration_udf']

    concentration_art = get_latest_input_artifact(concentration_step, lims_id,
                                                  lims)

    if concentration_art:
        return concentration_art.udf.get(concentration_udf)
    else:
        return None


def get_library_size(sample_id: str, lims: Lims, size_steps: List[str], workflow: str) -> int:
    """Getting the udf Size (bp) that in fact is set on the aggregate qc librar validation step."""

    if workflow == 'TWIST':
        out_art = get_output_artifact(size_steps, sample_id, lims, last=True)
        if out_art:
            return out_art.udf.get('Size (bp)')

    return None



"""
    Add or update analysis results for samples from bioinfo_processed into bioinf_sample collection
"""
import logging
import click


from vogue.tools.cli_utils import add_doc as doc
from vogue.build.bioinfo_analysis import build_bioinfo_sample
from vogue.load.bioinfo_analysis import load_analysis

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LOG = logging.getLogger(__name__)


@click.command(
    "sample",
    short_help="Process stats and results from bioinfo process and load sample into the bioinfo_sample collection.",
)
@click.option(
    "-c", "--analysis-case", help="The case to retrieve samples and load into bioinfo_sample."
)
@click.option(
    "-a",
    "--load-all",
    is_flag=True,
    help="Load all samples for all cases within bioinfo_processed.",
)
@click.option("--dry", is_flag=True, help="Load from sample or not. (dry-run)")
@doc(
    f"""
    Load samples analysis results from bioinfo processed collection
    into bioinfo sample collection.
    """
)
@click.pass_context
def bioinfo_sample(ctx: click.Context, dry, analysis_case, load_all):
    adapter = ctx.obj["adapter"]

    if (not load_all and not analysis_case) or (load_all and analysis_case):
        LOG.error(
            "--load-all and --analysis-case are mutually exclusive and cannot be used together"
        )
        raise click.Abort()

    current_processed_analysis = adapter.bioinfo_processed(analysis_case)
    LOG.info(
        "Loading following samples to bioinfo_samples: %s",
        ", ".join(current_processed_analysis["samples"]),
    )

    for sample in current_processed_analysis["samples"]:
        sample_analysis = build_bioinfo_sample(
            analysis_dict=current_processed_analysis, process_case=True, sample_id=sample
        )
        load_res = load_analysis(
            adapter=adapter,
            lims_id=sample,
            processed=True,
            is_sample=True,
            dry_run=dry,
            analysis=sample_analysis,
        )
        if load_res:
            LOG.info("Sample %s is loaded into database", sample)
        else:
            LOG.warning("Loading failed for sample %s", sample)



import logging

from mongo_adapter import MongoAdapter
from pymongo.collection import Collection

from datetime import datetime as dt

LOG = logging.getLogger(__name__)


def check_dates(analysis_result, current_document):
    """Function to pop analysysis results from tne new analysis if the results are older than
    the current results in the database"""

    if current_document and current_document.get("mip_dna") and analysis_result.get("mip_dna"):
        try:
            if current_document["mip_dna"]["added"] > analysis_result["mip_dna"]["added"]:
                analysis_result.pop("mip_dna")
        except:
            LOG.error("Invalid or missing mip_dna analysis date. Skiping analysis")
            analysis_result.pop("mip_dna")

    return analysis_result


class VogueAdapter(MongoAdapter):
    def setup(self, db_name: str):
        """Setup connection to a database"""

        if self.client is None:
            raise SyntaxError("No client is available")
        self.db = self.client[db_name]
        self.db_name = db_name
        self.sample_collection: Collection = self.db.sample
        self.bioinfo_processed_collection = self.db.bioinfo_processed
        self.bioinfo_raw_collection = self.db.bioinfo_raw
        self.bioinfo_samples_collection = self.db.bioinfo_samples
        self.app_tag_collection = self.db.application_tag
        self.flowcell_collection = self.db.flowcell
        self.reagent_label_collection = self.db.reagent_label
        self.reagent_label_category_collection = self.db.reagent_label_category
        self.genotype_analysis_collection = self.db.genotype_analysis

        LOG.info("Use database %s.", db_name)

    def add_or_update_document(self, document_news: dict, collection):
        """Adds/updates a document in the database"""

        document_id = document_news["_id"]

        update_result = collection.update_one(
            {"_id": document_id}, {"$set": document_news}, upsert=True
        )

        if not update_result.raw_result["updatedExisting"]:
            collection.update_one({"_id": document_id}, {"$set": {"added": dt.today()}})
            LOG.info("Added document %s.", document_id)
        elif update_result.modified_count:
            collection.update_one({"_id": document_id}, {"$set": {"updated": dt.today()}})
            LOG.info("Updated document %s.", document_id)
        else:
            LOG.info("No updates for document %s.", document_id)

    def add_or_update_bioinfo_raw(self, analysis_result: dict):
        """Functionality to add or update analysis for unprocessed aka raw bioinfo stat"""
        case_id = analysis_result["_id"]
        # pop _id key to make pushing easier
        analysis_result.pop("_id")
        update_result = self.db.bioinfo_raw.find_one({"_id": case_id})

        if update_result is None:
            self.db.bioinfo_raw.update_one(
                {"_id": case_id},
                {"$set": {**analysis_result, **{"added": dt.today()}}},
                upsert=True,
            )
            LOG.info("Added analysis sample %s.", case_id)
        else:
            self.db.bioinfo_raw.update_one(
                {"_id": case_id},
                {"$set": {**analysis_result, **{"updated": dt.today()}}},
                upsert=True,
            )
            LOG.info("Updated analysis for sample %s.", case_id)

    def add_or_update_bioinfo_processed(self, analysis_result: dict):
        """Functionality to add or update analysis for processed bioinfo stat"""
        case_id = analysis_result["_id"]
        # pop _id key to make pushing easier
        analysis_result.pop("_id")
        update_result = self.db.bioinfo_processed.find_one({"_id": case_id})

        if update_result is None:
            self.db.bioinfo_processed.update_one(
                {"_id": case_id},
                {"$set": {**analysis_result, **{"added": dt.today()}}},
                upsert=True,
            )
            LOG.info("Added analysis sample %s.", case_id)
        else:
            self.db.bioinfo_processed.update_one(
                {"_id": case_id},
                {"$set": {**analysis_result, **{"updated": dt.today()}}},
                upsert=True,
            )
            LOG.info("Updated analysis for sample %s.", case_id)

    def add_or_update_bioinfo_samples(self, analysis_result: dict):
        """Functionality to add or update bioinfo analysis for sample level results"""
        lims_id = analysis_result["_id"]
        current_document = self.db.bioinfo_samples.find_one({"_id": lims_id})
        analysis_result = check_dates(analysis_result, current_document)

        update_result = self.db.bioinfo_samples.update_one(
            {"_id": lims_id}, {"$set": analysis_result}, upsert=True
        )

        if not update_result.raw_result["updatedExisting"]:
            self.db.bioinfo_samples.update_one({"_id": lims_id}, {"$set": {"added": dt.today()}})
            LOG.info("Added sample %s.", lims_id)
        elif update_result.modified_count:
            self.db.bioinfo_samples.update_one({"_id": lims_id}, {"$set": {"updated": dt.today()}})
            LOG.info("Updated sample %s.", lims_id)
        else:
            LOG.info("No updates for sample %s.", lims_id)

    def sample(self, lims_id):
        return self.sample_collection.find_one({"_id": lims_id})

    def flowcell(self, run_id):
        return self.flowcell_collection.find_one({"_id": run_id})

    def app_tag(self, tag):
        return self.app_tag_collection.find_one({"_id": tag})

    def delete_sample(self):
        return None

    def sample_analysis(self, analysis_id: str):
        """Functionality to get analyses results"""
        return self.bioinfo_samples_collection.find_one({"_id": analysis_id})

    def bioinfo_raw(self, analysis_id: str):
        """Functionality to get analyses results"""
        return self.bioinfo_raw_collection.find_one({"_id": analysis_id})

    def bioinfo_processed(self, analysis_id: str):
        """Functionality to get analyses results"""
        return self.bioinfo_processed_collection.find_one({"_id": analysis_id})

    def find_samples(self, query: dict) -> list:
        """Function to find samples in samples collection based on query"""
        samples = self.sample_collection.find(query)
        return list(samples)

    def samples_aggregate(self, pipe: list):
        """Function to make a aggregation on the sample colleciton"""
        return self.sample_collection.aggregate(pipe)

    def flowcells_aggregate(self, pipe: list):
        """Function to make a aggregation on the flowcell colleciton"""
        return self.flowcell_collection.aggregate(pipe)

    def bioinfo_samples_aggregate(self, pipe: list):
        """Function to make a aggregation on the sample analysis colleciton"""
        return self.bioinfo_samples_collection.aggregate(pipe)

    def sample_collection_ids(self):
        pipe = [{"$project": {"_id": 1}}, {"$group": {"_id": None, "ids": {"$addToSet": "$_id"}}}]
        return self.sample_collection.aggregate(pipe)

    def genotype_analysis_aggregate(self, pipe: list):
        """Function to make a aggregation on the genotype analysis colleciton"""
        return self.genotype_analysis_collection.aggregate(pipe)

    def find_genotype_plate(self, plate_id: str):
        """find all samples from plate"""
        return self.genotype_analysis_collection.find({"plate": plate_id})

    def get_category(self, app_tag):
        """Function get category based on application tag from the application tag collection"""
        tag = self.app_tag_collection.find_one({"_id": app_tag}, {"category": 1})
        return tag.get("category") if tag else None

    def get_reagent_label_category(self, reagent_label):
        """Function get category based on application tag from the application tag collection"""
        category = self.app_tag_collection.find_one({"name": reagent_label}, {"category": 1})
        return category.get("category") if category else None

    def reagent_label_aggregate(self, pipe: list):
        """Function to make a aggregation on the reagent_label analysis colleciton"""
        return self.reagent_label_collection.aggregate(pipe)

    def get_all_reagent_label_names_grouped_by_category(self):
        """Function get all reagent label names grouped by category
        from the reagent_label_category colleciton"""
        pipe = [
            {"$group": {"_id": {"category": "$category"}, "reagent_labels": {"$push": "$name"}}}
        ]
        return self.reagent_label_category_collection.aggregate(pipe)

    def get_reagent_label_categories(self):
        """Function to get all categories from label_category_collection"""
        pipe = [
            {"$project": {"category": 1}},
            {"$group": {"_id": None, "categories": {"$addToSet": "$category"}}},
        ]
        reagent_label_categories = self.reagent_label_category_collection.aggregate(pipe)
        return list(reagent_label_categories)[0]["categories"]



from genologics.entities import Process
from genologics.lims import Lims

from vogue.parse.build.flowcell import run_data, filter_none
import datetime as dt


def build_run(run: Process, instrument: str, date: str) -> dict:
    """Build flowcell document from lims data."""

    mongo_run = {
        '_id': run.udf.get('Run ID'),
        'instrument': instrument,
        'date': dt.datetime.strptime(date, '%y%m%d')
    }

    for key, val in run.udf.items():
        if isinstance(val, dt.date):
            val = dt.datetime.strptime(val.isoformat(), '%Y-%m-%d')
        mongo_run[key] = val

    lane_data, avg_data = run_data(run)
    mongo_run['avg'] = avg_data
    mongo_run['lanes'] = lane_data

    mongo_run = filter_none(mongo_run)

    return mongo_run



import logging
import click
from vogue.load.reagent_label_category import load_all
from datetime import date, timedelta

from vogue.constants.constants import RUN_TYPES

LOG = logging.getLogger(__name__)


@click.command("reagent_label_categories", short_help="Load reagent_labels into db.")
@click.option(
    "-c",
    "--categories",
    multiple=True,
    type=str,
    help="Loads all reagent_label cathegories from lims.",
)
@click.pass_context
def reagent_label_categories(ctx: click.Context, categories: list = None):
    """Read and load reagent_label categories from lims"""
    adapter = ctx.obj["adapter"]
    lims = ctx.obj["lims"]
    if not lims:
        LOG.warning("Lims connection failed.")
        raise click.Abort()
    lims = lims
    load_all(adapter=adapter, lims=lims, categories=categories)



#!/usr/bin/env python

from fastapi import APIRouter, Request
from fastapi.responses import RedirectResponse
from urllib.parse import urlsplit, parse_qs, urlencode

router = APIRouter()


@router.get("/redirect")
async def update(request: Request, year: int):
    url = request.headers.get("referer")
    query = urlsplit(url).query
    params = parse_qs(query)
    params.update({"year": [year]})
    params_url = urlencode(params, doseq=True)

    return RedirectResponse(f"{url.split('?')[0]}?{params_url}")



#!/usr/bin/env python


def get_qc(adapter, year: int, metric_path: str) -> dict:
    """Prepares data for covid scatter plot"""

    metric = metric_path.split(".")[1]

    pipe = [
        {"$match": {"microsalt": {"$exists": "True"}}},
        {
            "$lookup": {
                "from": "sample",
                "localField": "_id",
                "foreignField": "_id",
                "as": "sample_info",
            }
        },
        {"$unwind": {"path": "$sample_info"}},
        {
            "$match": {
                "sample_info.category": {"$eq": "cov"},
                "sample_info.prepared_date": {"$exists": "True"},
                "sample_info.prep_method": {"$exists": "True"},
            }
        },
        {
            "$project": {
                "year": {"$year": "$sample_info.prepared_date"},
                "prep_method": "$sample_info.prep_method",
                "date": "$sample_info.prepared_date",
                metric: f"$microsalt.{metric_path}",
            }
        },
        {"$match": {"year": {"$eq": int(year)}}},
        {
            "$group": {
                "_id": {"prep_method": "$prep_method"},
                "data": {"$push": {metric: f"${metric}", "date": "$date", "id": "$_id"}},
            }
        },
    ]

    aggregate_result = adapter.bioinfo_samples_aggregate(pipe)
    results = {}
    for group_data in aggregate_result:
        group = group_data["_id"]["prep_method"]
        results[group] = []
        for sample in group_data["data"]:
            results[group].append([sample["date"], sample[metric], sample["id"]])

    return results



#!/usr/bin/env python

from vogue.constants.constants import THIS_YEAR
from vogue import __version__
from vogue.settings import templates

from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter
from fastapi import APIRouter, Depends, Request
from vogue.crud.find_plots.genotype import get_genotype_plate

router = APIRouter()


@router.get("/Bioinfo/Genotype/plate")
async def genotype_plate(
    request: Request,
    plate_id: str = "ID81",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):

    plot_data = get_genotype_plate(adapter, plate_id=plate_id)
    plot_data["plates"].sort()
    return templates.TemplateResponse(
        "genotype_plate.html",
        context=dict(
            request=request,
            plate_data=plot_data["data"],
            x_labels=plot_data["x_labels"],
            y_labels=plot_data["y_labels"],
            year_of_interest=year,
            header="Genotype",
            page_id="genotype_plate",
            version=__version__,
            plate_id=plot_data["plate_id"],
            plates=plot_data["plates"],
        ),
    )



import logging
import copy
import collections

from datetime import datetime as dt
import vogue.models.bioinfo_analysis as analysis_model
from vogue.tools.cli_utils import recursive_default_dict
from vogue.tools.cli_utils import convert_defaultdict_to_regular_dict

LOG = logging.getLogger(__name__)


def get_common_keys(valid_analysis: list, analysis_type: str):
    '''
    Match a list of values with keys from a MODEL dictionary

    input: valid_analysis as list
    output: analysis_common_keys as list
    '''

    # Match valid_analysis with the analysis_type of ANALYSIS_SETS
    analysis_common_keys = list()
    for my_analysis in valid_analysis:
        if my_analysis in list(
                analysis_model.ANALYSIS_SETS[analysis_type].keys()):
            analysis_common_keys.append(my_analysis)

    return analysis_common_keys


def extract_valid_analysis(analysis_dict: dict, analysis_type: str,
                           valid_analysis: list):
    '''
    Extracts analysis dictionary based on input analysis_dict. This function will remove analysis json that are not part
    of the matching model. analysis_type is a single key matching ANALYSIS_SETS's first level keys.

    Input:
        analysis_dict: A dictionary of bioinfo analysis stats.
        analysis_type: A string of analysis type. This is provided by user.
        valid_analysis: A list of analysis to be extracted from analysis dict.
    Output:
        analysis: A dictionary of valid_analysis as keys extracted from analysis_dict
    '''

    case_analysis_type = analysis_dict['case_analysis_type']

    analysis_common_keys = get_common_keys(valid_analysis, analysis_type)

    # A new dictionary is constructed instead of dropping unrelevant keys. Or maybe one could deepcopy analysis_dict and
    # remove the irrelevant keys.
    analysis = dict()

    for common_key in analysis_common_keys:
        # detect if multiqc analysis_dict is multiqc
        if case_analysis_type == "multiqc":
            analysis[common_key] = analysis_dict["multiqc"][
                "report_saved_raw_data"][common_key]
        elif case_analysis_type == "microsalt":
            analysis[common_key] = analysis_dict["microsalt"][common_key]

    return analysis


def build_processed_case(analysis_dict: dict,
                         analysis_type: str,
                         valid_analysis: list,
                         cleanup=False):
    '''
    Builds an analysis dict from input information provided by user.

    Input:
        analysis_dict: A dictionary of bioinfo stats to be prepared for bioinfo_processed collection
        analysis_type: A string for analysis_type to be extracted from from analysis_dict
        valid_analysis: A list of valid analysis to found within analysis_dict
        cleanup: Flag to cleanup unwanted keys from analysis_dict using info from valid_analysis and analysis_type
    Output:
        case_analysis: A dictionary with information about workflow and case_analysis_type(e.g. multiqc),
            workflow version, and date added.
    '''

    case_analysis = copy.deepcopy(analysis_dict)
    case_analysis_type = analysis_dict['case_analysis_type']
    analysis_workflow = analysis_dict['workflow']
    workflow_version = analysis_dict['workflow_version']
    case_analysis[analysis_workflow] = dict()
    case_analysis[analysis_workflow][case_analysis_type] = dict()
    if not cleanup:
        case_analysis[analysis_workflow][case_analysis_type] = analysis_dict[
            case_analysis_type]
    else:
        for my_analysis in analysis_model.ANALYSIS_DESC.keys():
            tmp_analysis_dict = extract_valid_analysis(
                analysis_dict=analysis_dict,
                analysis_type=my_analysis,
                valid_analysis=valid_analysis)
            if tmp_analysis_dict:
                case_analysis[analysis_workflow][case_analysis_type] = {
                    **case_analysis[analysis_workflow][case_analysis_type],
                    **tmp_analysis_dict
                }

    case_analysis[analysis_workflow][case_analysis_type] = {
        **case_analysis[analysis_workflow][case_analysis_type],
        **{
            'workflow_version': workflow_version,
            'added': dt.today()
        }
    }

    ## TODO: add the following to control over which valid analysis to load.
    ## Example:
    ##    elif cleanup and 'all' in analysis_type:
    ##    else:
    ##        case_analysis = extract_valid_analysis(analysis_dict=analysis_dict,
    ##                                               analysis_type=my_analysis,
    ##                                               valid_analysis=valid_analysis)

    return case_analysis


def build_unprocessed_case(analysis_dict: dict):
    '''
    Prepare a case analysis dictionary
    '''
    case_analysis = recursive_default_dict()
    case_analysis_type = analysis_dict['case_analysis_type']
    case_analysis = copy.deepcopy(analysis_dict[case_analysis_type])
    case_analysis = convert_defaultdict_to_regular_dict(case_analysis)

    return case_analysis


def build_mongo_case(analysis_dict: dict, case_analysis: dict,
                     processed=False):
    '''
    Build a mongo case document dictionary
    '''
    analysis_case = analysis_dict['case']
    analysis_workflow = analysis_dict['workflow']
    case_analysis_type = analysis_dict['case_analysis_type']
    workflow_version = analysis_dict['workflow_version']
    analysis_samples = list(analysis_dict['sample'])

    analysis = recursive_default_dict()

    if not processed:
        analysis[analysis_workflow][case_analysis_type] = list()
    else:
        analysis[analysis_workflow][case_analysis_type] = dict()

    analysis['workflows'] = list()
    analysis['case_analysis_types'] = list()
    analysis['samples'] = list()

    workflow_data = {
        **case_analysis,
        **{
            'workflow_version': workflow_version,
            'added': dt.today()
        }
    }
    if not processed:
        analysis[analysis_workflow][case_analysis_type].append(workflow_data)
    else:
        analysis[analysis_workflow][case_analysis_type] = workflow_data[
            analysis_workflow][case_analysis_type]
        #analysis[analysis_workflow] = {**analysis[analysis_workflow], **workflow_data[analysis_workflow][case_analysis_type]}

    analysis['case_analysis_types'].append(case_analysis_type)
    analysis['workflows'].append(analysis_workflow)
    analysis['samples'].extend(analysis_samples)

    analysis = convert_defaultdict_to_regular_dict(analysis)

    return analysis


def update_mongo_doc_case(mongo_doc: dict, analysis_dict: dict,
                          new_analysis: dict):
    '''
    Args:
        mongo_doc: an existing analysis retrieved from MongoDB 
        analysis_dict: a dictionary parsed from CLI
        new_analysis: new analysis dictionary to be loaded to MongoDB

    Returns:
        mongo_doc: an updated mongo_doc from Args

    Add or update mongo document for case data
    Adds or updates within processed or raw bioinfo collection
    '''

    analysis_samples = analysis_dict['sample']
    analysis_case = analysis_dict['case']
    analysis_type = analysis_dict['case_analysis_type']
    analysis_workflow_type = analysis_dict['workflow']
    workflow_version = analysis_dict['workflow_version']

    if analysis_workflow_type not in mongo_doc:
        mongo_doc[analysis_workflow_type] = {}

    if analysis_type not in mongo_doc[analysis_workflow_type]:
        mongo_doc[analysis_workflow_type][analysis_type] = new_analysis[
            analysis_workflow_type][analysis_type]
    else:
        mongo_doc[analysis_workflow_type] = {
            **new_analysis[analysis_workflow_type],
            **mongo_doc[analysis_workflow_type]
        }

    if analysis_workflow_type not in mongo_doc['workflows']:
        mongo_doc['workflows'].append(analysis_workflow_type)

    if analysis_type not in mongo_doc['case_analysis_types']:
        LOG.info("A new analysis type %s is added to the case %s",
                 analysis_type, analysis_case)
        mongo_doc['case_analysis_types'].append(analysis_type)

    for sample in analysis_samples:
        if sample not in mongo_doc['samples']:
            LOG.info("A new sample %s is added to the case %s", sample,
                     analysis_case)
            mongo_doc['samples'].append(sample)

    return mongo_doc


def build_analysis(analysis_dict: dict,
                   analysis_type: str,
                   valid_analysis: list,
                   current_analysis: dict,
                   process_case=False,
                   cleanup=False):
    '''
    Builds analysis dictionary based on input analysis_dict and prepares a mongo_doc.
    
    If not process_case, then do not validate any keys in the analysis_dict.
    This will only load into bioinfo_raw.

    If process_case, then extract valid keys from analysis_dict.
    '''

    analysis_case = analysis_dict['case']

    if not process_case:
        case_analysis = build_unprocessed_case(analysis_dict=analysis_dict)
    else:
        case_analysis = build_processed_case(analysis_dict=analysis_dict,
                                             analysis_type=analysis_type,
                                             valid_analysis=valid_analysis,
                                             cleanup=cleanup)

    analysis = build_mongo_case(analysis_dict=analysis_dict,
                                processed=process_case,
                                case_analysis=case_analysis)

    if current_analysis is None:
        # if there is no current analysis, return the built analysis
        mongo_doc = copy.deepcopy(analysis)
        mongo_doc['_id'] = analysis_case

    else:
        # if there is a current analysis, pop added and updated keys
        # and continue building the rest of the mongo sample
        mongo_doc = copy.deepcopy(current_analysis)
        mongo_doc.pop('added')
        if 'updated' in mongo_doc.keys():
            mongo_doc.pop('updated')

        mongo_doc = update_mongo_doc_case(mongo_doc=mongo_doc,
                                          analysis_dict=analysis_dict,
                                          new_analysis=analysis)

    return mongo_doc


def build_bioinfo_sample(analysis_dict: dict,
                         sample_id: str,
                         process_case=False):
    '''
    Builds sample analysis from analysis_dict

    analysis_dict is a processed dictionary, i.e. from bioinfo_processed
    '''

    analysis_case = analysis_dict['_id']
    analysis_workflows = analysis_dict['workflows']
    case_analysis_types = analysis_dict['case_analysis_types']

    if not process_case:
        return None

    mongo_doc = recursive_default_dict()
    mongo_doc['_id'] = sample_id
    mongo_doc['workflows'] = analysis_workflows
    mongo_doc['case_analysis_types'] = case_analysis_types
    for workflow in analysis_workflows:
        for analysis_type in case_analysis_types:
            simple_analysis_dict = analysis_dict[workflow][analysis_type]
            # check if sample_id is in key
            for metric, samples in simple_analysis_dict.items():
                if isinstance(samples, dict):
                    for sample in samples.keys():
                        # split key by underscore
                        if sample_id in sample.split("_"):
                            mongo_doc[workflow][metric] = samples[sample]
                            mongo_doc[workflow]['case'] = analysis_case

    mongo_doc = convert_defaultdict_to_regular_dict(mongo_doc)
    return mongo_doc



import logging

LOG = logging.getLogger(__name__)


def load_analysis(adapter, lims_id, analysis, processed=False, is_sample=False, dry_run=False):
    """Load information for a bioinfo analysis"""

    if dry_run:
        existing_analysis = adapter.sample(lims_id)
        if existing_analysis:
            LOG.info("The sample exists in the database")
        else:
            LOG.info("The sample does not exist in the database")

        LOG.info("Analysis information for sample %s to add/update: \n %s", lims_id, analysis)
        return

    if not processed and not is_sample:
        adapter.add_or_update_bioinfo_raw(analysis)
        load_status = True
    elif processed and not is_sample:
        adapter.add_or_update_bioinfo_processed(analysis)
        load_status = True
    elif processed and is_sample:
        adapter.add_or_update_bioinfo_samples(analysis)
        load_status = True
    else:
        LOG.warning("No analysis was loaded into database")
        load_status = False

    return load_status



from vogue import __version__ as version
from setuptools import setup, find_packages

try:
    with open("requirements.txt", "r") as f:
        install_requires = [x.strip() for x in f.readlines()]
except IOError:
    install_requires = []

setup(
    name="vogue",
    version=version,
    url="https://github.com/Clinical-Genomics/vogue",
    author="Maya Brandi",
    author_email="maya.brandi@scilifelab.se",
    install_requires=install_requires,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "console_scripts": ["vogue=vogue.commands:cli"],
    },
)



from vogue.adapter import VogueAdapter
from vogue.models.database import Sample, BioInfoSample
from vogue.crud.find import find_sample, find_bioinfo_sample
from vogue.exceptions import InsertError
from pymongo.results import InsertOneResult

import logging

LOG = logging.getLogger(__name__)


def create_sample(adapter: VogueAdapter, sample: Sample) -> str:
    """Function to load a new sample to the database."""

    if find_sample(adapter=adapter, sample_id=sample.id):
        raise InsertError(f"Sample {sample.id} already exist in sample collection.")

    result: InsertOneResult = adapter.sample_collection.insert_one(sample.dict(by_alias=True))
    LOG.info("Added sample document %s.", sample.id)

    return result.inserted_id


def create_bioinfo_sample(adapter: VogueAdapter, sample: BioInfoSample) -> str:
    """Function to load a new sample to the database."""

    if find_bioinfo_sample(adapter=adapter, sample_id=sample.id):
        raise InsertError(
            f"Bioinfo Sample {sample.id} already exist in bioinfo_samples collection."
        )

    result: InsertOneResult = adapter.bioinfo_samples_collection.insert_one(
        sample.dict(by_alias=True)
    )
    LOG.info("Added bioinfo sample document %s.", sample.id)

    return result.inserted_id



from vogue.constants.lims_constants import LANE_UDFS, MASTER_STEPS_UDFS
import numpy as np
import datetime as dt
import math


def filter_none(mongo_dict):
    """Function to filter out Nones and NaN from a dict."""
    for key in list(mongo_dict.keys()):
        val = mongo_dict[key]
        try:
            if val is None or math.isnan(val):
                mongo_dict.pop(key)
        except Exception:
            continue
    return mongo_dict


def run_data(run):
    """Function to get run info from lanes in a lims sequecing process.
    Reformates the data to be part of a document in the flowcell database.
    
    Arguments:
        run (Process): lims Process instance of sequencing type
    Returns: 
        lane_data (dict): run info per lane. 
            eq: {'Lane 1': {'% Aligned R2': 0.94, '% Bases >=Q30 R1': 90.67, '% Bases >=Q30 R2': 88.84,...}, 
                'Lane 2': {'% Aligned R2': 0.92, '% Bases >=Q30 R1': 91.67, '% Bases >=Q30 R2': 83.84,...}}
        avg_data (dict): average run info over all lanes. 
            eg: {'% Phasing R2': 0.09, '% Bases >=Q30': 89.755, ...}
        """
    lane_data = {}
    avg_data = {}
    if run.type.name in MASTER_STEPS_UDFS['sequenced']['nova_seq']:
        lanes = run.all_outputs()
    else:
        lanes = run.all_inputs()
    for lane in lanes:
        if not lane.location:
            continue
        if run.type.name in MASTER_STEPS_UDFS['sequenced']['nova_seq']:
            name = lane.name
            if not 'Lane' in name.split():
                continue
        else:
            name = lane.location[1]

        lane_data[name] = {}
        for udf in LANE_UDFS:
            value = lane.udf.get(udf)
            lane_data[name][udf] = value
            if not avg_data.get(udf):
                avg_data[udf] = []
            if isinstance(value, int) or isinstance(value, float):
                avg_data[udf].append(value)
        lane_data[name] = filter_none(lane_data[name])

    for udf, values in avg_data.items():
        avg_data[udf] = round(np.mean(values), 2)

    q30_r1 = MASTER_STEPS_UDFS['sequenced']['q30r1_udf']
    q30_r2 = MASTER_STEPS_UDFS['sequenced']['q30r2_udf']
    if q30_r1 in avg_data.keys() and q30_r2 in avg_data.keys():
        Q30 = np.mean([avg_data[q30_r1], avg_data[q30_r2]])
        avg_data['% Bases >=Q30'] = Q30
    avg_data = filter_none(avg_data)

    return lane_data, avg_data



def get_latest_analysis(case, analysis_type):
    """Get the latest analysis of anaöysis_type from one case"""

    all_analysis = case.get(analysis_type)
    if all_analysis:
        latest_analysis = all_analysis[0]
        for analysis in all_analysis:
            # analysis allways contains the key 'added'
            if analysis.get('added') > latest_analysis.get('added'):
                latest_analysis = analysis
        return latest_analysis
    return {}


def reduce_keys(dict_long_keys):
    """Cut keys generated in the multiqc report. 
    First entry is allways lims sample ID"""

    new_dict = {key.split('_')[0]: val for key, val in dict_long_keys.items()}
    return new_dict


class uSalt():
    """Class to prepare uSalt case_analysis results 
    for uSalt results in the sample_analysis collection"""

    def __init__(self, project):
        self.project = project
        self.uSalt_analysis = get_latest_analysis(project, 'microsalt')
        self.added = None
        self._set_init()

    def _set_init(self):
        if self.uSalt_analysis:
            self.added = self.uSalt_analysis.get('added')
            self.results = self.uSalt_analysis.get('results', {})

    def build_uSalt_sample(self, sample_id):
        """Bulding the uSalt analysis for one sample. 
        Returns {} if the date 'added' is empty."""

        if not self.added:
            return {}

        return {
            'results': self.results.get(sample_id),
            'added': self.added,
            'project': self.project['_id']
        }


class Mip_dna():
    """Class to prepare mip case_analysis results 
    for mip_dna results in the sample_analysis collection"""

    def __init__(self, case):
        self.case = case
        self.mip_dna_analysis = get_latest_analysis(case, 'mip-dna')
        self.added = None
        self.report_saved_raw_data = {}
        self.multiqc_picard_insertSize = {}
        self.multiqc_picard_HsMetrics = {}
        self._set_init()

    def _set_init(self):
        if self.mip_dna_analysis:
            self.added = self.mip_dna_analysis.get('added')
            self.report_saved_raw_data = self.mip_dna_analysis['multiqc'][
                'report_saved_raw_data']
            self.multiqc_picard_insertSize = reduce_keys(
                self.report_saved_raw_data['multiqc_picard_insertSize'])
            self.multiqc_picard_HsMetrics = reduce_keys(
                self.report_saved_raw_data['multiqc_picard_HsMetrics'])

    def build_mip_dna_sample(self, sample_id):
        """Bulding the mip analysis for one sample. 
        Returns {} if the date 'added' is empty."""

        if not self.added:
            return {}

        return {
            'multiqc_picard_insertSize':
            self.multiqc_picard_insertSize.get(sample_id),
            'multiqc_picard_HsMetrics':
            self.multiqc_picard_HsMetrics.get(sample_id),
            'added':
            self.added,
            'case':
            self.case['_id']
        }
